@@ -16,6 +16,8 @@
 
 #include "system.h"
 
+static thread_gc *free_thread_gcs;
+
 /* locally defined functions */
 #ifdef PTHREADS
 static s_thread_rv_t start_thread(void *tc);
@@ -33,7 +35,15 @@ void S_thread_init(void) {
     S_tc_mutex.owner = s_thread_self();
     S_tc_mutex.count = 0;
     s_thread_cond_init(&S_collect_cond);
-    S_tc_mutex_depth = 0;
+    s_thread_cond_init(&S_collect_thread0_cond);
+    s_thread_mutex_init(&S_alloc_mutex.pmutex);
+    s_thread_cond_init(&S_terminated_cond);
+    S_alloc_mutex.owner = 0;
+    S_alloc_mutex.count = 0;
+
+# ifdef IMPLICIT_ATOMIC_AS_EXPLICIT
+    s_thread_mutex_init(&S_implicit_mutex);
+# endif
 #endif /* PTHREADS */
   }
 }
@@ -41,30 +51,62 @@ void S_thread_init(void) {
 /* this needs to be reworked.  currently, S_create_thread_object is
    called from main to create the base thread, from fork_thread when
    there is already an active current thread, and from S_activate_thread
-   when there is no current thread.  we have to avoid thread-local
-   allocation in at least the latter case, so we call vector_in and
-   cons_in and arrange for S_thread to use find_room rather than
-   thread_find_room.  scheme.c does part of the initialization of the
+   when there is no current thread.  scheme.c does part of the initialization of the
    base thread (e.g., parameters, current input/output ports) in one
    or more places. */
 ptr S_create_thread_object(const char *who, ptr p_tc) {
   ptr thread, tc;
+  thread_gc *tgc;
   INT i;
 
-  tc_mutex_acquire()
+  tc_mutex_acquire();
 
   if (S_threads == Snil) {
-    tc = (ptr)S_G.thread_context;
+    tc = TO_PTR(S_G.thread_context);
+    tgc = &S_G.main_thread_gc;
+    GCDATA(tc) = TO_PTR(tgc);
+    tgc->tc = tc;
   } else { /* clone parent */
     ptr p_v = PARAMETERS(p_tc);
     iptr i, n = Svector_length(p_v);
-   /* use S_vector_in to avoid thread-local allocation */
-    ptr v = S_vector_in(space_new, 0, n);
+    ptr v;
+
+    tc = TO_PTR(malloc(size_tc));
+    if (free_thread_gcs) {
+      tgc = free_thread_gcs;
+      free_thread_gcs = tgc->next;
+    } else
+      tgc = malloc(sizeof(thread_gc));
 
-    tc = (ptr)malloc(size_tc);
     if (tc == (ptr)0)
       S_error(who, "unable to malloc thread data structure");
-    memcpy((void *)tc, (void *)p_tc, size_tc);
+    memcpy(TO_VOIDP(tc), TO_VOIDP(p_tc), size_tc);
+
+    GCDATA(tc) = TO_PTR(tgc);
+    tgc->tc = tc;
+
+    {
+      IGEN g; ISPC s;
+      for (g = 0; g <= static_generation; g++) {
+        for (s = 0; s <= max_real_space; s++) {
+          tgc->base_loc[g][s] = (ptr)0;
+          tgc->next_loc[g][s] = (ptr)0;
+          tgc->bytes_left[g][s] = 0;
+          tgc->sweep_loc[g][s] = (ptr)0;
+          tgc->sweep_next[g][s] = NULL;
+        }
+        tgc->bitmask_overhead[g] = 0;
+      }
+    }
+
+    tgc->during_alloc = 0;
+    tgc->pending_ephemerons = (ptr)0;
+    for (i = 0; i < (int)DIRTY_SEGMENT_LISTS; i++)
+      tgc->dirty_segments[i] = NULL;
+    tgc->queued_fire = 0;
+    tgc->preserve_ownership = 0;
+
+    v = S_vector_in(tc, space_new, 0, n);
 
     for (i = 0; i < n; i += 1)
       INITVECTIT(v, i) = Svector_ref(p_v, i);
@@ -73,22 +115,31 @@ ptr S_create_thread_object(const char *who, ptr p_tc) {
     CODERANGESTOFLUSH(tc) = Snil;
   }
 
- /* override nonclonable tc fields */
+  tgc->sweeper = main_sweeper_index;
+
+  /* override nonclonable tc fields */
   THREADNO(tc) = S_G.threadno;
   S_G.threadno = S_add(S_G.threadno, FIX(1));
 
   CCHAIN(tc) = Snil;
 
   WINDERS(tc) = Snil;
+  ATTACHMENTS(tc) = Snil;
+  HANDLERSTACK(tc) = Sfalse;
+  CACHEDFRAME(tc) = Sfalse;
   STACKLINK(tc) = SYMVAL(S_G.null_continuation_id);
   STACKCACHE(tc) = Snil;
 
  /* S_reset_scheme_stack initializes stack, size, esp, and sfp */
   S_reset_scheme_stack(tc, stack_slop);
-  FRAME(tc,0) = (ptr)&CODEIT(S_G.dummy_code_object,size_rp_header);
+  FRAME(tc,0) = TO_PTR(&CODEIT(S_G.dummy_code_object,size_rp_header));
 
  /* S_reset_allocation_pointer initializes ap and eap */
+  alloc_mutex_acquire();
   S_reset_allocation_pointer(tc);
+  alloc_mutex_release();
+  S_maybe_fire_collector(tgc);
+
   RANDOMSEED(tc) = most_positive_fixnum < 0xffffffff ? most_positive_fixnum : 0xffffffff;
   X(tc) = Y(tc) = U(tc) = V(tc) = W(tc) = FIX(0);
 
@@ -107,11 +158,9 @@ ptr S_create_thread_object(const char *who, ptr p_tc) {
 
   DSTBV(tc) = SRCBV(tc) = Sfalse;
 
- /* S_thread had better not do thread-local allocation */
   thread = S_thread(tc);
 
- /* use S_cons_in to avoid thread-local allocation */
-  S_threads = S_cons_in(space_new, 0, thread, S_threads);
+  S_threads = S_cons_in(tc, space_new, 0, thread, S_threads);
   S_nthreads += 1;
   SETSYMVAL(S_G.active_threads_id,
    FIX(UNFIX(SYMVAL(S_G.active_threads_id)) + 1));
@@ -124,9 +173,15 @@ ptr S_create_thread_object(const char *who, ptr p_tc) {
 
   GUARDIANENTRIES(tc) = Snil;
 
-  LZ4OUTBUFFER(tc) = NULL;
+  LZ4OUTBUFFER(tc) = 0;
 
-  tc_mutex_release()
+  CP(tc) = 0;
+
+  /* if a collection is needed, then ask the new thread to check right away */
+  if (Sboolean_value(S_symbol_value(S_G.collect_request_pending_id)))
+    TRAP(tc) = (ptr)1;
+
+  tc_mutex_release();
 
   return thread;
 }
@@ -138,9 +193,9 @@ IBOOL Sactivate_thread(void) { /* create or reactivate current thread */
   if (tc == (ptr)0) { /* thread created by someone else */
     ptr thread;
 
-   /* borrow base thread for now */
-    thread = S_create_thread_object("Sactivate_thread", S_G.thread_context);
-    s_thread_setspecific(S_tc_key, (ptr)THREADTC(thread));
+   /* borrow base thread to clone */
+    thread = S_create_thread_object("Sactivate_thread", TO_PTR(S_G.thread_context));
+    s_thread_setspecific(S_tc_key, TO_VOIDP(THREADTC(thread)));
     return 1;
   } else {
     reactivate_thread(tc)
@@ -193,7 +248,7 @@ static IBOOL destroy_thread(ptr tc) {
   ptr *ls; IBOOL status;
 
   status = 0;
-  tc_mutex_acquire()
+  tc_mutex_acquire();
   ls = &S_threads;
   while (*ls != Snil) {
     ptr thread = Scar(*ls);
@@ -201,8 +256,24 @@ static IBOOL destroy_thread(ptr tc) {
       *ls = Scdr(*ls);
       S_nthreads -= 1;
 
+      alloc_mutex_acquire();
+
      /* process remembered set before dropping allocation area */
-      S_scan_dirty((ptr **)EAP(tc), (ptr **)REAL_EAP(tc));
+      S_scan_dirty((ptr *)TO_VOIDP(EAP(tc)), (ptr *)TO_VOIDP(REAL_EAP(tc)));
+
+     /* close off thread-local allocation */
+      S_thread_start_code_write(tc, static_generation, 0, NULL, 0);
+      {
+        ISPC s; IGEN g;
+        thread_gc *tgc = THREAD_GC(tc);
+        for (g = 0; g <= static_generation; g++)
+          for (s = 0; s <= max_real_space; s++)
+            if (tgc->next_loc[g][s])
+              S_close_off_thread_local_segment(tc, s, g);
+      }
+      S_thread_end_code_write(tc, static_generation, 0, NULL, 0);
+
+      alloc_mutex_release();
 
      /* process guardian entries */
       {
@@ -211,7 +282,7 @@ static IBOOL destroy_thread(ptr tc) {
 	for (ges = GUARDIANENTRIES(tc); ges != Snil; ges = next) {
 	  obj = GUARDIANOBJ(ges);
 	  next = GUARDIANNEXT(ges);
-	  if (!IMMEDIATE(obj) && (si = MaybeSegInfo(ptr_get_segment(obj))) != NULL && si->generation != static_generation) {
+	  if (!FIXMEDIATE(obj) && (si = MaybeSegInfo(ptr_get_segment(obj))) != NULL && si->generation != static_generation) {
 	    INITGUARDIANNEXT(ges) = target;
 	    target = ges;
 	  }
@@ -226,20 +297,34 @@ static IBOOL destroy_thread(ptr tc) {
         if (Sboolean_value(SYMVAL(S_G.collect_request_pending_id))
             && SYMVAL(S_G.active_threads_id) == FIX(0)) {
           s_thread_cond_signal(&S_collect_cond);
+          s_thread_cond_signal(&S_collect_thread0_cond);
         }
       }
 
-      if (LZ4OUTBUFFER(tc) != NULL) free(LZ4OUTBUFFER(tc));
-      if (SIGNALINTERRUPTQUEUE(tc) != NULL) free(SIGNALINTERRUPTQUEUE(tc));
+      if (LZ4OUTBUFFER(tc) != (ptr)0) free(TO_VOIDP(LZ4OUTBUFFER(tc)));
+      if (SIGNALINTERRUPTQUEUE(tc) != (ptr)0) free(TO_VOIDP(SIGNALINTERRUPTQUEUE(tc)));
 
-      free((void *)tc);
+      if (THREAD_GC(tc)->preserve_ownership)
+        --S_num_preserve_ownership_threads;
+
+      /* Never free a thread_gc, since it may be recorded in a segment
+         as the segment's creator. Recycle manually, instead. */
+      THREAD_GC(tc)->sweeper = main_sweeper_index;
+      THREAD_GC(tc)->tc = (ptr)0;
+      THREAD_GC(tc)->next = free_thread_gcs;
+      free_thread_gcs = THREAD_GC(tc);
+
+      free(TO_VOIDP(tc));
+      
       THREADTC(thread) = 0; /* mark it dead */
       status = 1;
+
+      s_thread_cond_broadcast(&S_terminated_cond);
       break;
     }
     ls = &Scdr(*ls);
   }
-  tc_mutex_release()
+  tc_mutex_release();
   return status;
 }
 
@@ -251,7 +336,7 @@ ptr S_fork_thread(ptr thunk) {
   thread = S_create_thread_object("fork-thread", get_thread_context());
   CP(THREADTC(thread)) = thunk;
 
-  if ((status = s_thread_create(start_thread, (void *)THREADTC(thread))) != 0) {
+  if ((status = s_thread_create(start_thread, TO_VOIDP(THREADTC(thread)))) != 0) {
     destroy_thread((ptr)THREADTC(thread));
     S_error1("fork-thread", "failed: ~a", S_strerror(status));
   }
@@ -259,10 +344,10 @@ ptr S_fork_thread(ptr thunk) {
   return thread;
 }
 
-static s_thread_rv_t start_thread(p) void *p; {
-  ptr tc = (ptr)p; ptr cp;
+static s_thread_rv_t start_thread(void *p) {
+  ptr tc = TO_PTR(p); ptr cp;
 
-  s_thread_setspecific(S_tc_key, tc);
+  s_thread_setspecific(S_tc_key, TO_VOIDP(tc));
 
   cp = CP(tc);
   CP(tc) = Svoid; /* should hold calling code object, which we don't have */
@@ -274,13 +359,13 @@ static s_thread_rv_t start_thread(p) void *p; {
 
  /* find and destroy our thread */
   destroy_thread(tc);
-  s_thread_setspecific(S_tc_key, (ptr)0);
+  s_thread_setspecific(S_tc_key, NULL);
 
   s_thread_return;
 }
 
 
-scheme_mutex_t *S_make_mutex() {
+ptr S_make_mutex(void) {
   scheme_mutex_t *m;
 
   m = (scheme_mutex_t *)malloc(sizeof(scheme_mutex_t));
@@ -291,7 +376,7 @@ scheme_mutex_t *S_make_mutex() {
   m->owner = s_thread_self();
   m->count = 0;
 
-  return m;
+  return TO_PTR(m);
 }
 
 void S_mutex_free(scheme_mutex_t *m) {
@@ -299,32 +384,32 @@ void S_mutex_free(scheme_mutex_t *m) {
   free(m);
 }
 
-void S_mutex_acquire(scheme_mutex_t *m) {
+void S_mutex_acquire(scheme_mutex_t *m) NO_THREAD_SANITIZE {
   s_thread_t self = s_thread_self();
   iptr count;
   INT status;
 
   if ((count = m->count) > 0 && s_thread_equal(m->owner, self)) {
     if (count == most_positive_fixnum)
-      S_error1("mutex-acquire", "recursion limit exceeded for ~s", m);
+      S_error1("mutex-acquire", "recursion limit exceeded for ~s", TO_PTR(m));
     m->count = count + 1;
     return;
   }
-    
+
   if ((status = s_thread_mutex_lock(&m->pmutex)) != 0)
     S_error1("mutex-acquire", "failed: ~a", S_strerror(status));
   m->owner = self;
   m->count = 1;
 }
 
-INT S_mutex_tryacquire(scheme_mutex_t *m) {
+INT S_mutex_tryacquire(scheme_mutex_t *m) NO_THREAD_SANITIZE {
   s_thread_t self = s_thread_self();
   iptr count;
   INT status;
 
   if ((count = m->count) > 0 && s_thread_equal(m->owner, self)) {
     if (count == most_positive_fixnum)
-      S_error1("mutex-acquire", "recursion limit exceeded for ~s", m);
+      S_error1("mutex-acquire", "recursion limit exceeded for ~s", TO_PTR(m));
     m->count = count + 1;
     return 0;
   }
@@ -339,17 +424,24 @@ INT S_mutex_tryacquire(scheme_mutex_t *m) {
   return status;
 }
 
-void S_mutex_release(scheme_mutex_t *m) {
+IBOOL S_mutex_is_owner(scheme_mutex_t *m) NO_THREAD_SANITIZE {
+  s_thread_t self = s_thread_self();
+  return ((m->count > 0) && s_thread_equal(m->owner, self));
+}
+
+void S_mutex_release(scheme_mutex_t *m) NO_THREAD_SANITIZE {
   s_thread_t self = s_thread_self();
   iptr count;
   INT status;
 
   if ((count = m->count) == 0 || !s_thread_equal(m->owner, self))
-    S_error1("mutex-release", "thread does not own mutex ~s", m);
+    S_error1("mutex-release", "thread does not own mutex ~s", TO_PTR(m));
 
-  if ((m->count = count - 1) == 0)
+  if ((m->count = count - 1) == 0) {
+    m->owner = 0; /* needed for a memory model like ARM, for example */
     if ((status = s_thread_mutex_unlock(&m->pmutex)) != 0)
       S_error1("mutex-release", "failed: ~a", S_strerror(status));
+  }
 }
 
 s_thread_cond_t *S_make_condition() {
@@ -415,7 +507,7 @@ static inline int s_thread_cond_timedwait(s_thread_cond_t *cond, s_thread_mutex_
 
 #endif /* FEATURE_WINDOWS */
 
-#define Srecord_ref(x,i) (((ptr *)((uptr)(x)+record_data_disp))[i])
+#define Srecord_ref(x,i) (((ptr *)TO_VOIDP(((uptr)(x)+record_data_disp)))[i])
 
 IBOOL S_condition_wait(s_thread_cond_t *c, scheme_mutex_t *m, ptr t) {
   ptr tc = get_thread_context();
@@ -425,12 +517,14 @@ IBOOL S_condition_wait(s_thread_cond_t *c, scheme_mutex_t *m, ptr t) {
   I64 sec;
   long nsec;
   INT status;
+  IBOOL is_collect;
+  iptr collect_index = 0;
 
   if ((count = m->count) == 0 || !s_thread_equal(m->owner, self))
-    S_error1("condition-wait", "thread does not own mutex ~s", m);
+    S_error1("condition-wait", "thread does not own mutex ~s", TO_PTR(m));
 
   if (count != 1)
-    S_error1("condition-wait", "mutex ~s is recursively locked", m);
+    S_error1("condition-wait", "mutex ~s is recursively locked", TO_PTR(m));
 
   if (t != Sfalse) {
     /* Keep in sync with ts record in s/date.ss */
@@ -443,8 +537,30 @@ IBOOL S_condition_wait(s_thread_cond_t *c, scheme_mutex_t *m, ptr t) {
     nsec = 0;
   }
 
-  if (c == &S_collect_cond || DISABLECOUNT(tc) == 0) {
-    deactivate_thread(tc)
+  is_collect = (c == &S_collect_cond || c == &S_collect_thread0_cond);
+
+  if (is_collect) {
+    /* Remember the index where we record this tc, because a thread
+       might temporarily wait for collection, but then get woken
+       up (e.g., to make the main thread drive the collection) before
+       a collection actually happens. */
+    int i;
+    S_collect_waiting_threads++;
+    collect_index = maximum_parallel_collect_threads;
+    if (S_collect_waiting_threads <= maximum_parallel_collect_threads) {
+      /* look for an open slot in `S_collect_waiting_tcs` */
+      for (i = 0; i < maximum_parallel_collect_threads; i++) {
+        if (S_collect_waiting_tcs[i] == (ptr)0) {
+          collect_index = i;
+          S_collect_waiting_tcs[collect_index] = tc;
+          break;
+        }
+      }
+    }
+  }
+
+  if (is_collect || DISABLECOUNT(tc) == 0) {
+    deactivate_thread_signal_collect(tc, !is_collect)
   }
 
   m->count = 0;
@@ -453,10 +569,16 @@ IBOOL S_condition_wait(s_thread_cond_t *c, scheme_mutex_t *m, ptr t) {
   m->owner = self;
   m->count = 1;
 
-  if (c == &S_collect_cond || DISABLECOUNT(tc) == 0) {
+  if (is_collect || DISABLECOUNT(tc) == 0) {
     reactivate_thread(tc)
   }
 
+  if (is_collect) {
+    --S_collect_waiting_threads;
+    if (collect_index < maximum_parallel_collect_threads)
+      S_collect_waiting_tcs[collect_index] = (ptr)0;
+  }
+
   if (status == 0) {
     return 1;
   } else if (status == ETIMEDOUT) {
@@ -467,4 +589,3 @@ IBOOL S_condition_wait(s_thread_cond_t *c, scheme_mutex_t *m, ptr t) {
   }
 }
 #endif /* PTHREADS */
-
